# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/dictUtil.ipynb (unless otherwise specified).

__all__ = ['printDict', 'allKeysInDict', 'filterDt', 'stripDict', 'hashDict', 'saveDictToFile', 'loadDictFromFile',
           'saveStringToFile', 'loadStringFromFile', 'genSchema', 'printYaml', 'loadYaml', 'writeYaml', 'hash_dict',
           'deepUpdate']

# Cell
import yaml
from beartype import beartype

# Cell
def printDict(d:dict, length:int = 10, space = 0):
  '''
  print dictionary as first length value of values \n
  d: dict: the dictionary to be printed \n
  length: int: if the value is a string, then it will print the first length letter of the value, default = 10 \n
  space: int: the amount of additional space to be added for each nested dictionary, default = 0
  '''
  if type(d) != dict:
    print('this is not a dict')
    print(d)
  else:
    for k, v in d.items():
      if type(v) == dict:
        print(f"{' '*space}{k}")
        printDict(v, space=space+1)
      else:
        print(f"{' '*space}{k} : {v[:length] if type(v)==str else v}")

# Cell
def allKeysInDict(inputDict:dict, keys:list):
  '''
  checks whether all the keys given in the list are also keys in the inputDict dictionary \n
  inputDict: dict: the dictionary that will be used to check whether all the keys are in this dictionary \n
  keys: list: the list of keys
  '''
  return all(key in inputDict for key in keys)

# Cell
def filterDt(dtDict:dict):
  '''
  convert unjsonable datetime object to timestamp in the dictionary, this works for nested dictionary as well \n
  dtDict: dict: the dictionary inputted to convert the datetime object of its value
  '''
  from datetime import datetime
  return {k: (filterDt(v) if type(v) == dict else v) if type(v) != datetime else v.timestamp()
            for k,v in dtDict.items()}

# Cell
def stripDict(data:dict):
  '''
  if the value in the dictionary is a string, it will 'strip' the value to make it more clear \n
  data: dict: the dictionary inputted to be 'stripped'
  '''
  return {k: v.strip() if type(v) == str else v for k,v in data.items()}

# Cell
import hashlib, pickle, base64
def hashDict(data:dict, hasher= hashlib.sha1(), encoder = pickle.dumps):
  '''
  hashes the dictionary inputted \n
  data: dict: the dictionary inputted to be 'hashed'
  '''
  hasher.update(encoder(data))
  rawHash = hasher.digest()
  return base64.b64encode(rawHash).decode()

# Cell
import pickle
def saveDictToFile(data:dict, path:str):
  '''
  saves the dictionary to the file directed by the path \n
  data: dict: the dictionary to be saved \n
  path: str: the file path to the file the dictionary is going to be saved
  '''
  with open(path, 'wb')as f:
    pickle.dump(data,f,protocol=pickle.HIGHEST_PROTOCOL)

def loadDictFromFile(path:str):
  '''
  returns the dictionary that is saved in the file \n
  path: str: the path taken to the file of the dictionary
  '''
  with open(path, 'rb') as f:
    return pickle.load(f)

# Cell
def saveStringToFile(data:str, path:str):
  '''
  saves the string to the file directed by the path \n
  data: str: the string to be saved \n
  path: str: the file path to the file the string is going to be saved
  '''
  with open(path, 'w')as f:
    f.write(data)
def loadStringFromFile(path:str):
  '''
  returns the string that is saved in the file \n
  path: str: the path taken to the file of the string
  '''
  with open(path, 'r')as f:
    return f.read()

# Cell
def genSchema(inputDict:dict, format_='yaml')->(dict,str):
  '''
  generate a json schema from dict \n
  format: return schema in json or yaml, 'both' can be inputted to return a tuple of (json, yaml), output = dict or str, default = 'yaml' \n
  inputDict: dict: the dict inputted to be used to generate the schema
  '''
  from genson import SchemaBuilder
  import yaml
  builder = SchemaBuilder()
  builder.add_object(inputDict)
  schema = builder.to_schema()
  if format_=='yaml':
    return yaml.dump(schema)
  elif format_ == 'json':
    return schema
  elif format_ == 'both':
    return schema, yaml.dump(schema)
  else:
    return schema, yaml.dump(schema)

# Cell
@beartype
def printYaml(input_:dict, returnYaml = False)->(str, None):
  yamlStr:str =  yaml.dump(input_)
  print(yamlStr)
  if returnYaml:
    return yamlStr

# Cell
@beartype
def loadYaml(path:str, loader = yaml.FullLoader)->dict:
  '''load yaml string
    input:
      path:str: path of the data to load
    return:
      diet of the yaml

  '''
  with open(path) as f:
    r = yaml.load(f.read(), Loader = loader)
  return r

# Cell
@beartype
def writeYaml(path:str, data:dict):
  with open(path, 'w') as f:
    f.write(yaml.dump(data))


# Cell
import functools
import json
import ujson
import pickle
def hash_dict(func):
    """Transform mutable dictionnary
    Into immutable
    Useful to be compatible with cache
    """
    class HDict(dict):
        def __hash__(self):
          return hash(ujson.dumps(self, sort_keys=True))
#           return hash(tuple(frozenset(sorted(self.items()))))
#           print(self)
#           return hash(hashDict(self))
#             return hash(frozenset(self.items()))

    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        args = tuple([HDict(arg) if isinstance(arg, dict) else arg for arg in args])
        kwargs = {k: HDict(v) if isinstance(v, dict) else v for k, v in kwargs.items()}
        return func(*args, **kwargs)
    return wrapped

# Cell
import collections
def deepUpdate(d, u):
    for k, v in u.items():
        if isinstance(d, collections.Mapping):
            if isinstance(v, collections.Mapping):
                r = deepUpdate(d.get(k, {}), v)
                d[k] = r
            else:
                d[k] = u[k]
        else:
            d = {k: u[k]}
    return d


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/string.ipynb (unless otherwise specified).

__all__ = ['urlDecode', 'urlEncode']

# Cell
def urlDecode(s:str)->dict:
    """Decode a url-encoded string into a dictionary of key/value pairs"""
    return dict([x.split('=') for x in s.split('&')])

# Cell
def urlEncode(s:str)->str:
    """Encode a dictionary of key/value pairs into a url-encoded string"""
    return '&'.join([f'{k}={v}' for k,v in s.items()])


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/sentryUtil.ipynb (unless otherwise specified).

__all__ = ['logSentry', 'autologSentry']

# Cell
from sentry_sdk import add_breadcrumb, capture_exception, capture_message
from .exception import traceback
from typing import Any
from copy import deepcopy
from .exception import errorString
import functools, sentry_sdk, inspect

# Cell
def logSentry(message:str, data:Any = (lambda :{})(), level:str = 'info', section:str='main'):
  '''
    just add docs for ease of logging to sentry
    Input:
      message ::str:: required :: message to send to sentry
      data ::dict:: optional :: and object to send to sentry (default is an empty dict)
      level ::str::optional:: log level (default:info)
      section ::str::optional:: section name or function name (default: main)
    Response:
      Bool:: true means logged properly, false for error, print error message to console
  '''
  try:
    add_breadcrumb(
      category=section,
      data={'data':deepcopy(data)},
      level=level,
      message=message
    )
    return True
  except Exception as e:
    print(message, data, level, section)
    print(f'error is {e}, {traceback()}')
    return False

# Cell
def autologSentry(verbose=False):
  def outer(f):
    @functools.wraps(f)
    def inner(*args,**kwargs):
      try:
        r = f(*args,**kwargs)
        message = str(f.__name__)+str(inspect.signature(f))
        sentry_sdk.add_breadcrumb(category=f.__name__, message=message, data={
          'args': args, 'kwargs': kwargs, 'output':r }, level='info')
        if verbose: print(f.__name__, message, { 'args': args, 'kwargs': kwargs, 'output':r })
        return r
      except Exception as e:
        sentry_sdk.add_breadcrumb(category=f.__name__, message=errorString(), data={
          'args': args, 'kwargs': kwargs, 'error': str(e) }, level='critical')
        if verbose: print(f.__name__, errorString(), {'args': args, 'kwargs': kwargs, 'error':str(e)} )
        raise e
    return inner
  return outer


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/pdUtils.ipynb (unless otherwise specified).

__all__ = ['getDfHash', 'saveLocalCache', 'saveLocalHash', 'loadLocalCache', 'loadLocalHash', 'saveRemoteHash',
           'saveRemoteCache', 'loadRemoteCache', 'loadRemoteHash', 'PandasDataFrameAttribute', 'forceType']

# Cell
import pandas as pd
from hashlib import sha1
from .dictUtil import saveStringToFile, loadStringFromFile
from pynamodb.constants import BINARY
from pynamodb.attributes import Attribute, UnicodeAttribute
from pynamodb.models import Model
from beartype import beartype
from s3bz.s3bz import S3
import os, logging

# Cell
from io import BytesIO
from typing import Callable

@beartype
def getDfHash(df:pd.DataFrame,
              hashingAlgorithm: Callable = lambda x: sha1(x).hexdigest())->str:
  '''
    get a hash of a pandas dataframe
    this uses sha1 algorithm
    inputs:
      df: pd.DataFrame: a pandas dataframe
      hashingAlgoritm: callable: a hasing function which takes bytes input
    response:
      string hash
  '''
  f:BytesIO = BytesIO()
  df.to_feather(f)
  return hashingAlgorithm(f.read())


# Cell
def saveLocalCache( data:pd.DataFrame, path:str = '/tmp/cache',
                   saveHash:bool = True, force:bool = True):
  '''
    save cache of the dataframe to local location
    data:pd.DataFrame: dataframe to save
    path: str: path to save cache
    saveHash: bool: whether to save the hash digest
  '''
  ##check cache
  if not force:
    localHash = loadLocalHash(f'{path}.hash')
    dataHash = getDfHash(data)
    if dataHash == localHash :
      logging.debug('hash is the latest, skipping')
      return True
  ##save hash
  if saveHash:
    saveLocalHash(data, path=f'{path}.hash')
  # save cache
  logging.debug('saving cache')
  r =  data.to_feather(path)
  return r

def saveLocalHash( data:pd.DataFrame, path = '/tmp/cache.hash', force = False):
  dataHash = getDfHash(data)
  return saveStringToFile(dataHash,path)

def loadLocalCache( path = '/tmp/cache', throw = True):
  if not os.path.exists(path):
    if throw:
      raise Exception('cache doesnt exist')
  return pd.read_feather(path)
def loadLocalHash( path = '/tmp/cache.hash'):
  if not os.path.exists(path): raise Exception('hash doesnt exist')
  return loadStringFromFile(path)

# Cell
def saveRemoteHash(data:pd.DataFrame, key='', bucket='', **kwargs):
  hashKey = f'{key}-hash'
  hashString = getDfHash(data)
  dictToSave= {'hash': hashString }
  print(f'hashKey is {hashKey}')
  print('saving hash to s3')
  S3.save(key=hashKey,objectToSave=dictToSave, bucket=bucket, **kwargs )
  print(f'saved hash {hashString}')


def saveRemoteCache(data:pd.DataFrame, key = '',
                    bucket = '', localCachePath='/tmp/cache', localHashPath='/tmp/hash', **kwargs):

  saveLocalCache(data=data, path = localCachePath)
  saveLocalHash(data=data, path = localHashPath)
  saveRemoteHash(data=data, key = key, bucket=bucket)
  S3.saveFile(key=key, path=localCachePath, bucket=bucket, **kwargs)

def loadRemoteCache(key='', bucket='', **kwargs):
  path = '/tmp/tmpPath'
  S3.loadFile(key,path=path ,bucket=bucket, **kwargs)
  df = pd.read_feather(path)
  return df

def loadRemoteHash(key='', bucket='', **kwargs):
  hashKey = f'{key}-hash'
  print(f'loading hashkey {hashKey}')
  loadedHash= S3.load(hashKey,bucket=bucket, **kwargs).get('hash')
  print(f'loaded hash is {loadedHash}')
  return loadedHash

# Cell
class PandasDataFrameAttribute(Attribute):
  attr_type = BINARY
  def serialize(self, value: pd.DataFrame)->bin:
    bio = BytesIO()
    value.to_feather(bio)
    data:bin = bio.getvalue()
    return data
  def deserialize(self, value: bin)->pd.DataFrame:
    bio = BytesIo(bin)
    df: pd.DataFrame = pd.read_feather(bio)
    return df

# Cell
# class PandasSeriesAttribute(Attribute):
#   attr_type = BINARY
#   def serialize(self, value: pd.Series)->bin:
#     bio = BytesIO()
#     df = s.to_frame()
#     value.to_feather(bio)
#     data:bin = bio.getvalue()
#     return data
#   def deserialize(self, value: bin)->pd.DataFrame:
#     bio = BytesIo(bin)
#     df: pd.DataFrame = pd.read_feather(bio)
#     return df

# Cell
from .schema import getTypes

# Cell
def forceType(url:str, df:pd.DataFrame, defaultType=str)->pd.DataFrame:
  typeDict = getTypes(url)
  typeList = {col:typeDict.get(col, str) for col in df.columns}
  print(typeList)
  df = df.astype(typeList)
  print(df.dtypes)
  return df


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/dataclassUtil.ipynb (unless otherwise specified).

__all__ = ['enforce_types']

# Cell
import inspect
import typing
from contextlib import suppress
from functools import wraps

# Cell
def enforce_types(callable):
    spec = inspect.getfullargspec(callable)

    def check_types(*args, **kwargs):
        parameters = dict(zip(spec.args, args))
        parameters.update(kwargs)
        for name, value in parameters.items():
            with suppress(KeyError):  # Assume un-annotated parameters can be any type
                type_hint = spec.annotations[name]
                if isinstance(type_hint, typing._SpecialForm):
                    # No check for typing.Any, typing.Union, typing.ClassVar (without parameters)
                    continue
                try:
                    actual_type = type_hint.__origin__
                except AttributeError:
                    # In case of non-typing types (such as <class 'int'>, for instance)
                    actual_type = type_hint
                # In Python 3.8 one would replace the try/except with
                # actual_type = typing.get_origin(type_hint) or type_hint
                if isinstance(actual_type, typing._SpecialForm):
                    # case of typing.Union[…] or typing.ClassVar[…]
                    actual_type = type_hint.__args__

                if not isinstance(value, actual_type):
                    raise TypeError('Unexpected type for \'{}\' (expected {} but found {})'.format(name, type_hint, type(value)))

    def decorate(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            check_types(*args, **kwargs)
            return func(*args, **kwargs)
        return wrapper

    if inspect.isclass(callable):
        callable.__init__ = decorate(callable.__init__)
        return callable

    return decorate(callable)




# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/encrypt.ipynb (unless otherwise specified).

__all__ = ['hashSha256', 'encryptAes', 'encryptPasswordAes', 'decryptAes', 'decryptPasswordAes']

# Cell
from Crypto.Cipher import AES
from beartype import beartype
from typing import Tuple, Optional
from hashlib import sha256

# Cell
def hashSha256(data:bytes, salt:Optional[bytes] = b'')->bytes:
  m=sha256()
  m.update(data)
  m.update(salt)
  return m.digest()

# Cell
@beartype
def encryptAes(key:bytes, data:bytes)-> Tuple[bytes,bytes]:
  '''
    note that key has to be 16 bits long
    response:
      cipherText in bytes
  '''
  cipher = AES.new(key, AES.MODE_EAX)
  ciphertext, tag = cipher.encrypt_and_digest(data)
  return ciphertext, cipher.nonce


# Cell
@beartype
def encryptPasswordAes(password:bytes, data:bytes) ->Tuple[bytes,bytes]:
  key = hashSha256(password)[:16]
  return encryptAes(key=key,data=data)

# Cell
@beartype
def decryptAes(key:bytes, ciphertext:bytes, nonce:bytes)->bytes:
  cipher = AES.new(key, AES.MODE_EAX, nonce=nonce)
  plaintext = cipher.decrypt(ciphertext)
  return plaintext

# Cell
@beartype
def decryptPasswordAes(ciphertext:bytes, password:bytes, nonce:bytes)->bytes:
  key = hashSha256(password)[:16]
  return decryptAes(key, ciphertext, nonce)
